#atlas-gap-ana, v. 1.63, aleksandra jekic
#use pattern "python atlas-gap-ana.py 2021-02-03T1400"
#to analyze one hour of ripe atlas data
#use pattern "python atlas-gap-ana.py 2021-02-03"
#to analyze 24 hours of ripe atlas data

import ujson as json
import urllib.request
import bz2, math
import geoip2.database, pygeoip
import sys, os
#import cProfile #Used for profiling time usage
import sqlite3

#Filepaths used by the program

CityDatabase = "/usr/share/GeoIP/GeoLite2-City.mmdb"
ASDatabaseIPv4 = "/usr/share/GeoIP/GeoIPASNum.dat"
ASDatabaseIPv6 = "/usr/share/GeoIP/GeoIPASNumv6.dat"
LogPath = "log/" #The folder where you want the completed new logs stored
DBPath = "" #Where you want the temporary database stored


#Creates the URL for the log file to download based on request

def getLogName(logsNotAnalyzed):

    day = logsNotAnalyzed[0:10]

    link = "https://data-store.ripe.net/datasets/atlas-daily-dumps/" + day \
        + "/traceroute-" + logsNotAnalyzed + ".bz2"

    return link



#Checks the source address of the last set of traceroute replies.
#Traceroute is defined as successful if one of these
#packages is from the destination address.

#Defined as failed if 1) the traceroute does not
#contain a destination (DNS failure etc.), 2) the
#traceroute did not receive any replies or 3) the
#last set of ICMP replies is not from the
#destination address of the traceroute.


def checkSuccess(data):

    try:

        destination = data["dst_addr"]
        lastResult = data["result"][-1]["result"]

    except:
        return False

    for packet in lastResult:
        try:

            if packet["from"] == destination:
                return True

        except:
            pass

    return False



#Checks for source and destination address.
#Marks log entries without it which are DNS failures etc.
#These log entries are dropped by the function calling this one.
#Otherwise returns address in format "srcaddress/destaddress".

def getUniquePair(data):

    try:

        dst = str(data["prb_id"]) + "/" + data["dst_addr"]

        if data["from"] == "" or data["dst_addr"] == "" or str(data["prb_id"]) == "":
            dst = None

    except:
        dst = None

    return dst



#Checks if the probe failed to even start the traceroute
#Typical errors are "name resolution failed: non-recoverable 
#failure in name resolution" and "connect failed: Network is unreachable"

def checkForProbeFail(data, uniquePair, probeFailCount):

    if "error" in data["result"][0]:
        probeFailCount += 1
        return None, probeFailCount
    
    else:
        return uniquePair, probeFailCount



#Translates error codes into their corresponding textual meaning.
#Based on various RFC and documentation from RIPE Atlas
#https://atlas.ripe.net/docs/data_struct/#v5000_traceroute

def explainError(name):

    if name == "N":
        return "Network unreachable"
    elif name == "H":
        return "Destination unreachable"
    elif name == "A":
        return "Administratively prohibited"
    elif name == "P":
        return "Protocol unreachable"
    elif name == "p":
        return "Port unreachable"
    elif name == 6:
        return "IPv6: Route to destination rejected"
    elif name == 10:
        return "IPv4: Administratively prohibited"
    elif name == 5:
        return "IPv6: Source address not allowed"
    elif name == "h":
        return "Beyond scope"
    elif name == 9:
        return "IPv4: Destination network administratively prohibited"
    else:
        return(name)



#Checks the final set of 3 ICMP replies from
#each traceroute, collecting the error messages

def checkForErrors(data):

    errors = []
    dups = 0
    lates = 0


    #Gets the final set of ICMP replies

    lastResult = data["result"][-1]["result"]


    #Now collects errors from each reply in the set

    for packet in lastResult:

        packetErrors = []


        #If a reply is late, it's discarded
        #since we can't tell what hop it is from (?)
        #We still choose to count it though

        if "late" in packet:
            lates = lates + 1
            continue


        #I believe dupes may just add confusion
        #when looking at the output of the program,
        #so I've chosen to discard those as well
        #We still choose to count it though

        if "dup" in packet:
            dups = dups + 1
            continue

        
        #If everything is all right so far, the error
        #message of the given reply is recorded

        try:
            packetErrors.append(explainError(packet["err"]))

        except:
            pass


        #If there was no error message, checks for other conditions

        if len(packetErrors) == 0:

            #Checks if we received a "No reply" in the format 'x':'*'

            if "x" in packet:
                packetErrors.append("No reply")

            #Check if there was a failure in the router
            
            elif "error" in packet:
                packetErrors.append(packet["error"])

            #If no explanation has yet been found, then the
            #traceroute is labeled as incomplete

            else:
                packetErrors.append("Incomplete")
            
        errors.append(packetErrors)

        #Sometimes the last set of ICMP replies is something like 5 times 'x':'*'.
        #This may be mostly lates and dupes from unknown hops.
        #This is commented out here, so that we get each reply

        #if len(errors) > 3:
        #    errors = [['Incomplete'], ['Incomplete'], ['Incomplete']]

    return lates, dups, errors



#Checks sets of ICMP replies from the last until it finds
#one that is not lost so we know which router replied last

def findLastReplyIP(data):

    #Grabs the set of last ICMP replies received

    result = data["result"][-1]["result"]

    notfound = True
    x = -1

    while notfound:

        #Searches each packet

        for packet in result:
            try:
                return packet["from"]
            except:
                pass

        
        #If an IP is not found, we jump one set of answers back
        #If we run out of sets of ICMP replies, then
        #there were no IPs in any of the replies received.
        #This does really occur sometimes.

        x = x - 1

        try:
            result = data["result"][x]["result"]
        except:
            return "No IP found in results"



#Returns all the RTT information from a traceroute log

def collectRtt(data):

    rtt = []

    for x in range(0, len(data["result"])):

        rtt.append([])

        for packet in data["result"][x]["result"]:
            
            try:
                rtt[x].append(packet["rtt"])
                continue
            except:
                pass
                
            if "late" in packet:
                rtt[x].append("Late")
            else:
                rtt[x].append("Unknown")

    return rtt



#Initializes storage for information about a route,
#defined as a unique pair of source and destination.

def initialize(data, memory, uniquePair, marker):

    memory[uniquePair] = {}
    memory[uniquePair][0] = {}
    memory[uniquePair][0]["ip_ver"] = data["af"]
    memory[uniquePair][0]["timeStamps"] = data["timestamp"]
    memory[uniquePair][0]["routeStatus"] = checkSuccess(data)
    memory[uniquePair][0]["from"] = data["from"]
    memory[uniquePair][0]["topology"] = marker
    memory[uniquePair][0]["rtt"] = marker

    if memory[uniquePair][0]["routeStatus"]:
        memory[uniquePair][0]["errors"] = "No error"
        memory[uniquePair][0]["dups"] = 0
        memory[uniquePair][0]["lates"] = 0
        memory[uniquePair][0]["lastReply"] = "Destination IP"
    else:
        lates, dups, errors = checkForErrors(data)
        memory[uniquePair][0]["errors"] = errors
        memory[uniquePair][0]["dups"] = dups
        memory[uniquePair][0]["lates"] = lates
        memory[uniquePair][0]["lastReply"] = findLastReplyIP(data)



#Since RIPE Atlas logs are not sorted by time,
#we must do the sorting ourselves to determine outages

def sortMemoryByTimestamps(data, memory, uniquePair):

    for x in range(len(memory[uniquePair])-1, -1, -1):

        if data["timestamp"] < memory[uniquePair][x]["timeStamps"]:

            temp = memory[uniquePair][x]
            
            memory[uniquePair][x] = memory[uniquePair][x+1]
            memory[uniquePair][x+1] = temp



#Counts the amount of outages for a unique route
#If there are no successful traceroutes,
#then a series of failures is not counted as an outage

def countOutages(memory, uniquePair):

    currentState = memory[uniquePair][0]["routeStatus"]
    outage = not currentState
    posfound = False

    if outage:
        outages = 1
    else:
        outages = 0

    for x in range(1, len(memory[uniquePair]), 1):

        if currentState != memory[uniquePair][x]["routeStatus"]:
            outage = not memory[uniquePair][x]["routeStatus"]
            currentState = memory[uniquePair][x]["routeStatus"]

            if outage == True:
                outages += 1
            else:
                posfound = True
    
    if not posfound:
        outages = 0

    return outages



#Retrieves and parses RTT information from SQL
#database using marker that was saved

def getRtt(marker, cursor):

    k = str(cursor.execute("SELECT rtt FROM rtt WHERE key = ?", (marker,),).fetchone())
    k = k[2:-3]
    k = json.loads(k)
    
    return k



#Creates information about all outages for a given route.
#A lot of stuff happens here, so it can be a bit overwhelming to read.

def analyzeEachOutage(memory, uniquePair, cursor):

    outages = countOutages(memory, uniquePair)
    index = 0 #Used to track position in the traceroutes as they are read for outages
    newmemory = {} #We'll store everything here and drop the old memory file to save RAM.
    totalOutageDuration = 0 #Used to track total duration of all outages for a route.
    incomplete = 0 #Used to track total amount of incomplete traceroutes in a route

    for r in range(0, outages):

        newmemory[r] = {}

        markers = [] #Saing the SQLlite ids of failed routes so they can be retrieved
        #after we have used and deleted the successful routes.
        #They are not immediately retrieved to save RAM.
        failrtts = [] #Does the same for their RTTs.
        srcadr = [] #Collects src adresses of each traceroute in the outage in the very unlikely case they differ


        outageErrors = [] #Used to track errors per outage
        failures = 0 #Used to track amount of failed traceroutes in an outage
        frequency = 0 #Used to calculate polling frequency
        begins_in_other_hour = False
        end_in_other_hour = False
        freqchange = 0 #Counts amount of time differences recorded, used to find polling frequency

        
        #Skips through any successful traceroutes, since these are not part of outages

        while (memory[uniquePair][index]["routeStatus"] == True) and index < len(memory[uniquePair]):
            index = index + 1

        #If the status is false, this is recorded as the start of an outage.
        #All information needed is collected.

        start = memory[uniquePair][index]["timeStamps"]
        markers.append(memory[uniquePair][index]["topology"])
        failrtts.append(getRtt(memory[uniquePair][index]["rtt"], cursor))
        srcadr.append(memory[uniquePair][index]["from"])

        if index > 0:
            frequency = memory[uniquePair][index]["timeStamps"] - memory[uniquePair][index-1]["timeStamps"]
            freqchange += 1

        outageErrors.append(memory[uniquePair][index]["errors"])
        finalReply = memory[uniquePair][index]["lastReply"]
        failures = 1
        index = index + 1

        #If the log started on an outage we want to make a note of that

        if index == 0:
            begins_in_other_hour = True

        #Now tries to find the end of the outage (it may not have an end)

        notfound = True

        #As long the end has not been found and there are still traceroutes left to read we keep checking them.

        while notfound and (index in memory[uniquePair]):

            #While the outage is still ongoing, we keep adding information about it.

            if (memory[uniquePair][index]["routeStatus"] == False):
                outageErrors.append(memory[uniquePair][index]["errors"])
                markers.append(memory[uniquePair][index]["topology"])
                failures = failures + 1
                frequency = (memory[uniquePair][index]["timeStamps"] - memory[uniquePair][index-1]["timeStamps"]) + frequency
                freqchange += 1
                failrtts.append(getRtt(memory[uniquePair][index]["rtt"], cursor))
                srcadr.append(memory[uniquePair][index]["from"])
                index = index + 1

            #When the outage ends, we record information about this.

            elif (memory[uniquePair][index]["routeStatus"] == True):
                
                end = memory[uniquePair][index]["timeStamps"]
                frequency = (memory[uniquePair][index]["timeStamps"] - memory[uniquePair][index-1]["timeStamps"]) + frequency
                freqchange += 1
                frequency = frequency/freqchange #Calculates average polling
                duration = (end - start) #Calculates duration of each outage
                notfound = False


        #If we ran out of traceroutes to check while the outage was still ongoing,
        #we have to record the time it ends and the duration as unknown.

        if not (index in memory[uniquePair]):
            end_in_other_hour = True
            end = memory[uniquePair][index-1]["timeStamps"]
            duration = (end - start) #Calculates duration of each outage

        #We count the amount of incomplete traceroutes.

        for msg in outageErrors[0]:
            if str(msg) == "['Incomplete']":
                incomplete = incomplete + 1
                break

        #We add the duration from the last
        #outage to the total duration of outages.

        totalOutageDuration = totalOutageDuration + duration

        #In case there is no RTTs information for the failed traceroutes

        if failrtts == []:
            failrtts = "Unknown"

        #Finally we add all the information about the outage 
        #we're currently looking at to our new memory.

        newmemory[r]["startTime"] = start
        newmemory[r]["endTime"] = end
        newmemory[r]["failures"] = failures
        newmemory[r]["duration"] = duration
        newmemory[r]["fromfalse"] = srcadr
        newmemory[r]["lastReply"] = finalReply
        newmemory[r]["errors"] = outageErrors
        newmemory[r]["markers"] = markers
        newmemory[r]["polling_time"] = int(frequency)
        newmemory[r]["begins_in_other_hour"] = begins_in_other_hour
        newmemory[r]["end_in_other_hour"] = end_in_other_hour
        newmemory[r]["failrtts"] = failrtts

        #After adding the data for the final outage, we throw the accumulated data 
        #from all the outages in with the first outage where we can easily retrieve it later

        if r+1 == outages:

            newmemory[0]["incomplete"] = incomplete
            newmemory[0]["ip_ver"] = memory[uniquePair][0]["ip_ver"]
            newmemory[0]["total_outage_duration"] = totalOutageDuration
            newmemory[0]["length"] = len(memory[uniquePair])

            #We also need to get all the probe_ids before we delete the old memory
            #And get the successful route topologies from the SQLlite database with their RTT data
            #Additionally we want to get the total dup/late count

            newmemory[0]["from"] = []
            newmemory[0]["whole_routes"] = []
            newmemory[0]["total_dups"] = 0
            newmemory[0]["total_lates"] = 0
            newmemory[0]["rtt"] = []


            for x in memory[uniquePair]:

                newmemory[0]["from"].append(memory[uniquePair][x]["from"])
                newmemory[0]["total_dups"] = newmemory[0]["total_dups"] + memory[uniquePair][x]["dups"]
                newmemory[0]["total_lates"] = newmemory[0]["total_lates"] + memory[uniquePair][x]["lates"]
                
                if memory[uniquePair][x]["routeStatus"] == True:
                    k = str(cursor.execute("SELECT topology FROM topologies WHERE key = ?", (memory[uniquePair][x]["topology"],),).fetchone())
                    k = k[2:-3]
                    k = json.loads(k)
                    newmemory[0]["whole_routes"].append(k)
                    newmemory[0]["rtt"].append(getRtt(memory[uniquePair][x]["rtt"], cursor))

    return newmemory



#After we have used all the successful route topologies, we delete these
#With our RAM back, we can now get the topologies of the failed routes

def exchangeTopology(memory, cursor):

    for x in memory:
        del memory[x][0]["whole_routes"]
    
    for x in memory:
        for y in memory[x]:
            memory[x][y]["topology"] = []
            for z in memory[x][y]["markers"]:
                k = str(cursor.execute("SELECT topology FROM topologies WHERE key = ?", (z,),).fetchone())
                k = k[2:-3]
                k = json.loads(k)
                memory[x][y]["topology"].append(k)

    return memory



#Updates memory if a unique route has been seen before and sorts it by time

def updateMemory(data, memory, uniquePair, marker):

    memory[uniquePair][len(memory[uniquePair])] = {}
    memory[uniquePair][len(memory[uniquePair])-1]["timeStamps"] = data["timestamp"]
    memory[uniquePair][len(memory[uniquePair])-1]["routeStatus"] = checkSuccess(data)
    memory[uniquePair][len(memory[uniquePair])-1]["from"] = data["from"]
    memory[uniquePair][len(memory[uniquePair])-1]["ip_ver"] = data["af"]
    memory[uniquePair][len(memory[uniquePair])-1]["topology"] = marker
    memory[uniquePair][len(memory[uniquePair])-1]["rtt"] = marker

    if not checkSuccess(data):
        lates, dups, errors = checkForErrors(data)
        memory[uniquePair][len(memory[uniquePair])-1]["dups"] = dups
        memory[uniquePair][len(memory[uniquePair])-1]["lates"] = lates
        memory[uniquePair][len(memory[uniquePair])-1]["errors"] = errors
        memory[uniquePair][len(memory[uniquePair])-1]["lastReply"] = findLastReplyIP(data)
        
    else:
        memory[uniquePair][len(memory[uniquePair])-1]["errors"] = "No error"
        memory[uniquePair][len(memory[uniquePair])-1]["dups"] = 0
        memory[uniquePair][len(memory[uniquePair])-1]["lates"] = 0
        memory[uniquePair][len(memory[uniquePair])-1]["lastReply"] = "Destination IP"
        

    sortMemoryByTimestamps(data, memory, uniquePair)



#Prepares total log summary data for each hourly log
#so that it can be printed to a JSON file

def prepareTotalData(numberOfTraceroutes, successfulTraceroutes, failedTraceroutes, uniqueProbeCount, lowestTimestamp, highestTimestamp, probeFailCount):

    totalData = {
        "total_traceroutes": numberOfTraceroutes,
        "event_type": "log_summary",
        "start_time": lowestTimestamp,
        "end_time": highestTimestamp,
        "total_succeeded": successfulTraceroutes,
        "total_failed": failedTraceroutes,
        "failed_at_probe": probeFailCount,
        "total_probes": uniqueProbeCount,
        "version": "1.63 atlas-gap-ana aleksandra jekic"
    }

    return totalData



#Returns the highest RTT from a list

def findMaxRtt(rtt):

    highest = -1

    for y in rtt:
        if y != "Unknown" and y != "Late":
            if y > highest:
                highest = y

    if highest == -1:
        return "Unknown"
        
    return highest



#Returns the lowest RTT from a list

def findMinRtt(rtt):

    lowest = float('inf')

    for y in rtt:
        if y != "Unknown" and y != "Late":
            if y < lowest:
                lowest = y

    if lowest == float('inf'):
        return "Unknown"
        
    return lowest



#Formats the RTT data into a list std_dv()
#can use to find the standard deviation

def top_std_dev(rtt):

    rtts = []

    for y in rtt:
        if y != "Unknown" and y != "Late":
            rtts.append(y)

    return std_dv(rtts)



#Analyzes the raw route data for how frequently different IPs
#appear on different hops in different orders.
#Keeps track of RTTs information about hops
#Also finds the length of the shortest and longest route.

def analyzeTopology(topology, rtt):

    #First finds the max and min length of the routes

    listlen = [len(i) for i in topology]
    maxm = max(listlen)
    minm = min(listlen)

    #Makes a post for each possible hop
    #based on topology and RTT data

    charts = {}
    pos = -1

    for x in topology:
        pos = pos + 1
        hop = 0
        for y in x:
            counter = 0
            hops = []
            for z in y:
                item = {}
                item["hop"] = hop
                item["ip"] = z
                item["rtt"] = rtt[pos][hop][counter]
                counter += 1
                hops.append(item)
            if hop not in charts:
                charts[hop] = []
            charts[hop].append(hops)
            hop += 1

    #Merges duplicates and calculates how common hops are
    #Also finds max and min RTT for each hop and standard deviation

    newchart = []
    finalchart = []

    for x in charts:
        hop = {}
        hopcount = 0
        totcount = 0
        for y in charts[x]:
            for z in y:
                cur = z["ip"] + "hop" + str(z["hop"])
                if cur not in hop:
                    hop[cur] = {}
                    hop[cur]["hop"] = z["hop"]
                    hop[cur]["ip"] = z["ip"]
                    hop[cur]["rtt"] = [z["rtt"]]
                    hop[cur]["rtt_count"] = 1
                else:
                    hop[cur]["rtt"].append(z["rtt"])
                    hop[cur]["rtt_count"] += 1
                totcount += 1
        for z in hop:
            hop[z]["portion"] = str(round(hop[z]["rtt_count"]/totcount * 100)) + "%"
            hop[z]["maxrtt"] = findMaxRtt(hop[z]["rtt"])
            hop[z]["minrtt"] = findMinRtt(hop[z]["rtt"])
            hop[z]["rtt_std_dv"] = top_std_dev(hop[z]["rtt"])

            if hop[z]["ip"] != "*":
                del hop[z]["rtt"]
                finalchart.append(hop[z])

        hopcount += 1
        newchart.append(hop)

    return maxm, minm, finalchart



#Analyzes each outage and gets rid off the data we're not using
#I. e. routes with only one traceroute or no outages

def cleanUp(memory, cursor):

    newmemory = {}

    for x in memory.copy():

        if len(memory[x]) > 1:

            if countOutages(memory, x) > 0:

                newmemory[x] = analyzeEachOutage(memory, x, cursor)
    
    return newmemory



#Finds the total summed RTT of each traceroute

def sumRtt(data):

    sumRtt = []

    for x in data:
        sumd = 0
        for y in x:
            pair = 0
            count = 0
            for z in y:
                if z != "Unknown" and z != "Late":
                    pair += z
                    count += 1
            if pair > 0:
                sumd += pair/count
        sumRtt.append(round(sumd,2))

    return sumRtt



#Calculates standard deviation

def std_dv(rtt):

    if rtt == []:
        return "Unknown"

    count = 0
    
    for x in rtt:
        count += x

    count = count/len(rtt)

    ncount = 0

    for x in rtt:
        ncount += pow((x - count), 2)

    ncount = ncount/len(rtt)

    return round(math.sqrt(ncount), 2)



#Counts the amount of each error type from the end of a failed traceroute
#When mode is true, it can count errors from multiple traceroutes

def countErrorTypes(errors, mode):
    
    arr = {}

    if mode:
        for x in errors:
            for y in x:
                for z in y:
                    if str(z[0]) in arr:
                        arr[str(z[0])] += 1
                    else:
                        arr[str(z[0])] = 1
    else:
        for y in errors:
            for z in y:
                if str(z[0]) in arr:
                    arr[str(z[0])] += 1
                else:
                    arr[str(z[0])] = 1

    return arr



#Prepares an hourly summary of each unique route
#so that it can be printed to a JSON file

def preparePairData(memory):

    pairData = {}

    for x in memory:

        pairname = x.replace("/", "-to-")
        pair = x.split("/")

        allerrors = []
        failed = 0

        for y in memory[x]:
            allerrors.append(memory[x][y]["errors"])
            failed = failed + memory[x][y]["failures"]

        if memory[x][0]["whole_routes"] != []:
            maxm, minm, top = analyzeTopology(memory[x][0]["whole_routes"], memory[x][0]["rtt"])
        else:
            maxm = 0
            minm = 0
            top = "No successful routes known"
        
        if len(set(memory[x][0]["from"])) > 1:
            srcadr = memory[x][0]["from"]
        else:
            srcadr = memory[x][0]["from"][0]

        entry = {
        "pair_id": pairname,
        "event_type": "summary",
        "dst_addr": pair[1],
        "src_addr": srcadr,
        "prb_id": pair[0],
        "start_time": memory[x][0]["startTime"],
        "last_response": memory[x][len(memory[x])-1]["endTime"],
        "traceroutes_recorded": memory[x][0]["length"],
        "failed_traceroutes": str(round((failed/memory[x][0]["length"]) * 100)) + "%",
        "outage_count": len(memory[x]),
        "outage_at_log_start": memory[x][0]["begins_in_other_hour"],
        "outage_at_log_end": memory[x][len(memory[x])-1]["end_in_other_hour"],
        "incomplete_outages": memory[x][0]["incomplete"],
        "total_dupes": memory[x][0]["total_dups"],
        "total_late": memory[x][0]["total_lates"],
        "total_outage_errors": countErrorTypes(allerrors, True),
        "total_outage_duration_in_seconds": memory[x][0]["total_outage_duration"],
        "max_hop": maxm,
        "min_hop": minm,
        "topology": top
        }

        pairData[x] = entry

    return pairData



#Gets AS and location information from GeoIP2

def useGeoIP(lastReply, ipv):

    #In case the lookups fail and also so the program will still
    #otherwise work if it's not run on a server with GeoIP2 setup like this

    AS_number = "Unknown"
    AS_name = "Unknown"
    location = "Unknown"
    latitude = "Unknown"
    longitude = "Unknown"

    #Mostly doing try/exceptions here because it's faster
    #when the lookups are mostly successful

    try:

        with geoip2.database.Reader(CityDatabase) as reader:
            response = reader.city(lastReply)

            #In many cases there is no city name
            try:
                location = response.country.name + ", " + response.city.name
            except:
                location = response.country.name

            try:
                latitude = response.location.latitude
                longitude = response.location.longitude
            except:
                pass

    except:
        pass

    #AS Numbers from seperate database

    try:
        if ipv == 4:
            gi = pygeoip.GeoIP(ASDatabaseIPv4)
        elif ipv == 6:
            gi = pygeoip.GeoIP(ASDatabaseIPv6)

        result = gi.org_by_addr(lastReply)
        AS_number = result.split(' ', 1)[0]
        AS_name = result.split(' ', 1)[1]

    except:
        pass

    return AS_number, AS_name, location, latitude, longitude



#Attempts to identify a main cause for outage
#based on the last error information received

def classifyOutage(errors):

    cause = "Unknown"
    arr = {}

    #Counts each error type

    for x in errors:
        for y in x:
            if str(y) in arr:
                arr[str(y)] += 1
            else:
                arr[str(y)] = 1

    #Makes sure there are not more than
    #one "most common" error message

    if len(arr) == 1:
        key = next(iter(arr))
    else:
        key = max(arr, key=arr.get)
        count = 0
        for x in y:
            if x == arr[key]:
                count += 1
        if count > 1:
            return "Various causes"

    #Returns the correct description

    if key == "['No reply']":
        cause = "Stopped route"
    elif key == "['Incomplete']":
        cause = "Not enough information"
    else:
        cause = key[2:-2]

    return cause



#Prepares a detailed summary of each unique outage
#so that it can be printed to a JSON file

def prepareOutageData(memory):

    outageData = {}

    for y in memory:

        #Checks if there are any outages for this route
        #using data made by preparePairData()

        pairname = y.replace("/", "-to-")
        pair = y.split("/")
        ipv = memory[y][0]["ip_ver"] #Gets the IP version for use by GeoIP2
        collection = {} #Stores data for all outages in each route, added to outageData in the end
        
        
        for x in memory[y]:

            #Gets the min and max lengths of the failed routes
            listlen = [len(i) for i in memory[y][x]["topology"]]
            maxm = max(listlen)
            minm = min(listlen)
            
            #Gets the GeoIP2 information
            AS_number, AS_name, location, latitude, longitude = useGeoIP(memory[y][x]["lastReply"], ipv)

            maxm, minm, top = analyzeTopology(memory[y][x]["topology"], memory[y][x]["failrtts"])

            if len(set(memory[y][x]["fromfalse"])) > 1:
                srcadr = memory[y][x]["fromfalse"]
            else:
                srcadr = memory[y][x]["fromfalse"][0]

            outage = {
                "pair_id": pairname,
                "event_type": "outage",
                "order": x+1,
                "dst_addr": pair[1],
                "src_addr": srcadr,
                "prb_id": pair[0],
                "start_time": memory[y][x]["startTime"],
                "last_response": memory[y][x]["endTime"],
                "outage_at_log_start": memory[y][x]["begins_in_other_hour"],
                "outage_at_log_end": memory[y][x]["end_in_other_hour"],
                "failure_count": memory[y][x]["failures"],
                "polling_frequency": round(memory[y][x]["polling_time"]),
                "duration": memory[y][x]["duration"],
                "last_reply_from": memory[y][x]["lastReply"],
                "AS_number": AS_number,
                "AS_owner": AS_name,
                "location": location,
                "coordinates": str(latitude) + "," + str(longitude),
                "outage_errors": countErrorTypes(memory[y][x]["errors"], False),
                "error_type": classifyOutage(memory[y][x]["errors"]),
                "max_hop": maxm,
                "min_hop": minm,
                "topology": top
            }

            #Returns log without coordinates if latitude and longitude is unknown
            if latitude == "Unknown" or longitude == "Unknown":
                del outage["coordinates"]


            collection[x] = outage

        outageData[y] = collection
        

    return outageData



#Writes all information prepared to a JSON file

def writeToJSON(totalData, pairData, outageData, request):

    path = LogPath + str(request) + "-finished"

    with open(path, "w") as outfile:
        outfile.write(json.dumps(totalData) + "\n")
        for x in pairData:
            outfile.write(json.dumps(pairData[x]) + "\n")
            try:
                for y in outageData[x]:
                    outfile.write(json.dumps(outageData[x][y]) + "\n")
            except:
                pass



#Returns the route of a single traceroute

def printRoute(data):

    result = data["result"]
    arr = []

    for x in result:
        
        hop = []
        
        for y in x["result"]:
            if "from" in y:
                hop.append(y['from'])
            elif "x" in y:
                hop.append(y['x'])
            elif "error" in y:
                hop.append(y["error"])

        arr.append(hop)

    return arr

    

#Most important part of the program
#which analyzes the downloaded logs

def analyzeLog(request):

    #Initializes data
    numberOfTraceroutes = 0
    successfulTraceroutes = 0
    failedTraceroutes = 0
    uniqueProbeCount = 0
    uniqueProbeDictionary = {}
    current_probe_id = 0
    lowestTimestamp = float('inf')
    highestTimestamp = 0
    memory = {}
    pairData = {}
    outageData = {}
    marker = 0
    probeFailCount = 0


    link = getLogName(request) #Creates the url where the log data for the requested day exists
    req = urllib.request.Request(link)
    bz2log = urllib.request.urlopen(req) #Opens the url as a file

    #Connects to or creates a SQLlite3 database to store topology information
    #It has slow lookup, but topologies take a lot of space
    #yet we're actually not using that many of them so this is ideal

    connection = sqlite3.connect(DBPath + str(request) + ".db")
    cursor = connection.cursor()

    try:
        cursor.execute("DROP TABLE topologies")
        cursor.execute("DROP TABLE rtt")
        cursor.execute("CREATE TABLE topologies (key TEXT, topology TEXT)")
        cursor.execute("CREATE TABLE rtt (key TEXT, rtt TEXT)")
    except:
        cursor.execute("CREATE TABLE topologies (key TEXT, topology TEXT)")
        cursor.execute("CREATE TABLE rtt (key TEXT, rtt TEXT)")

    #Makes lookups much faster
    cursor.execute("CREATE UNIQUE INDEX idx_key1 ON topologies (key)")
    cursor.execute("CREATE UNIQUE INDEX idx_key2 ON rtt (key)")

    print("Started analyzing log:", request)

    #Loads each line of the hourly log from RIPE Atlas servers
    for line in bz2.BZ2File(bz2log, 'rb'):
        data = json.loads(line)

        #Counts total traceroutes
        numberOfTraceroutes += 1

        #Counts successful and failed traceroutes
        if checkSuccess(data):
            successfulTraceroutes +=1
        else:
            failedTraceroutes +=1


        #Checks if a probe has been counted before.
        current_probe_id = data["prb_id"]
        if current_probe_id not in uniqueProbeDictionary:
            uniqueProbeDictionary[current_probe_id] = 0
        
        
        #Keeps track of lowest and highest timestamp in each hourly log
        if (data["timestamp"] < lowestTimestamp):
            lowestTimestamp = data["timestamp"]

        if (data["timestamp"]) > highestTimestamp:
            highestTimestamp = data["timestamp"]

            
        #Gets identifier for route
        uniquePair = getUniquePair(data)

        #Makes sure that traceroutes that fail in the starting probe are
        #skipped by making their uniquePair identifier say "None"
        uniquePair, probeFailCount = checkForProbeFail(data, uniquePair, probeFailCount)

        if uniquePair is None:

            pass

        else:
            

            #Initializes the memory if we have never seen this route before
            if uniquePair not in memory:
                cursor.execute(('INSERT INTO topologies (key, topology) values (?, ?)'), (marker, json.dumps(printRoute(data))))
                cursor.execute(('INSERT INTO rtt (key, rtt) values (?, ?)'), (marker, json.dumps(collectRtt(data))))
                initialize(data, memory, uniquePair, marker)
                marker = marker + 1

            #Otherwise adds it to existing memory
            else:
                cursor.execute(('INSERT INTO topologies (key, topology) values (?, ?)'), (marker, json.dumps(printRoute(data))))
                cursor.execute(('INSERT INTO rtt (key, rtt) values (?, ?)'), (marker, json.dumps(collectRtt(data))))
                updateMemory(data, memory, uniquePair, marker)
                marker = marker + 1
    

    #Prepares final data for writing to JSON
    memory = cleanUp(memory, cursor) #Creates outage reports and deletes unused data
    pairData = preparePairData(memory) #Creates the JSON files for each unique pair report
    memory = exchangeTopology(memory, cursor) #Deletes successful route topologies and gets unsuccessful from SQLlite database
    outageData = prepareOutageData(memory) #Creates the JSON reports for each outage
    uniqueProbeCount = len(uniqueProbeDictionary)
    totalData = prepareTotalData(numberOfTraceroutes, successfulTraceroutes, 
    failedTraceroutes, uniqueProbeCount, lowestTimestamp, highestTimestamp, probeFailCount)
    
    #Finally writes all reports to JSON
    writeToJSON(totalData, pairData, outageData, request)
    #Deletes log file we've finished with
    os.remove(DBPath + str(request) + ".db")
    print("Completed analyzing:", request)



#Main program

def main():

    #Checks the nature of the request, sends
    #error message if it doesn't make sense

    try:
        request = str(sys.argv[1])
    except:
        print("Please execute program using pattern 'python atlas-gap-ana.py" +
        " 2021-03-03T1400'.")

    if len(request) == 15 and len(sys.argv) == 2:
        analyzeLog(request)

    else:
        print("Please execute program using pattern 'python atlas-gap-ana.py" +
        " 2021-03-03T1400'.")


main()